# Loonies 8192 r50 for DOS and SDL platforms
# Copyright (C) 2018-2020 Thomas Perl <m@thp.io>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


# Tile Editor + Color Scheme Generator
# ====================================
# Tiles are the digits for score display and the tiles
# for the actual gameplay. When run without any args,
# the tile editor is launched. When run with "--export"
# as the argument, it reads in all files + schemes and
# writes out "schemes.h" which includes the color schemes
# (see "selection" variable, as not all generated schemes
# are used in the final game) and the 2-bit packed pixel
# data for digits (10x10) and tiles (16x16).


import colorsys
import time
import numpy as np
import glob
import sys

blocksize = 16
is_small = False

# Which color schemes to actually export to the file
selection = """
blackwhite
bunt0
bunt1
similar0
blass0
similar4
dark0
bunt7
blass2
bunt8
similar6
funky0
similar7
bunt9
bunt13
funky7
bunt14
blass5
dark1
funky8
blass7
funky15
""".strip().split()

export = ('--export' in sys.argv)

colors = []
colorx = []
colorj = []
colorf1 = []
colorf2 = []
for i in range(0, 256, 4):
    v = float(i) / 255
    colors.append([int(x*255) for x in colorsys.hsv_to_rgb(v, 0.8, 0.9)])
    colorx.append([int(x*255) for x in colorsys.hsv_to_rgb(v, 0.2, 0.9)])
    colorj.append([int(x*255) for x in colorsys.hsv_to_rgb(v, 0.9, 0.3)])
    colorf1.append([int(x*255) for x in colorsys.hsv_to_rgb(v, 0.6, 0.9)])
    colorf2.append([int(x*255) for x in colorsys.hsv_to_rgb(v, 0.4, 0.7)])
lc = len(colors)

coffset = 1

schemes = []
schemenames = []

def load_digits():
    digits = []
    for i in range(10):
        pixels = np.frombuffer(open('digits/%d.dat' % (i,), 'rb').read(), dtype=np.uint8)
        if len(pixels) == 100:
            pixels.shape = (10, 10)
        else:
            pixels.shape = (blocksize, blocksize)
        digits.append(pixels.copy())
    return digits

digits = load_digits()

def draw_digit(x, y, digit, color=0):
    pixels = digits[ord(digit) - ord('0')]
    for yy in range(10):
        for xx in range(10):
            if pixels[yy][xx] != 0:
                scr.set_at((xx+x, yy+y), get_tile_color(pixels[yy][xx], color))


schemes.append([(90, 90, 90), (215, 215, 215), (40, 40, 40), (20, 20, 20)])
schemenames.append('blackwhite')

for i in range(16):
    o = i*4
    schemes.append([
        colors[o%lc],
        colors[int(o+30)%lc],
        colors[int(o+46)%lc],
        colors[int(o+50)%lc],
    ])
    schemenames.append('bunt%d' % i)

for i in range(8):
    o = i*8
    schemes.append([
        colors[o%lc],
        colorx[int(o+10)%lc],
        colorx[int(o+32)%lc],
        colorx[int(o+36)%lc],
    ])
    schemenames.append('blass%d' % i)

for i in range(4):
    o = i*16
    schemes.append([
        colors[(o+7)%lc],
        colorj[int(o+20)%lc],
        colorj[int(o+50)%lc],
        colorj[int(o+52)%lc],
    ])
    schemenames.append('dark%d' % i)

for i in range(16):
    o = i*4
    schemes.append([
        colorf1[o%lc],
        colorf2[int(o+8)%lc],
        colors[int(o+32)%lc],
        colors[int(o+42)%lc],
    ])
    schemenames.append('funky%d' % i)

for i in range(8):
    o = i*99 + 7
    schemes.append([
        colors[(o+4)%lc],
        colorj[int(o+8)%lc],
        colorj[int(o+12)%lc],
        colorj[int(o+16)%lc],
    ])
    schemenames.append('similar%d' % i)

existing = sorted(glob.glob('tiles/*.dat') + glob.glob('digits/*.dat'))
existidx = 0

def tile_file_to_2bit(filename):
    pixels = open(filename, 'rb').read()
    bits = ''
    for i in range(0, len(pixels), 4):
        value = (pixels[i] << 6) | (pixels[i+1] << 4) | (pixels[i+2] << 2) | pixels[i+3]
        bits += '0x%02x, ' % value
        if i % 20 == 16:
            bits = bits[:-1] + '\n      '
    return bits

if export:
    print('Exporting to schemes.h')
    with open('schemes.h', 'w') as fp:
        print('/* Generated from files in artwork/colors-and-tiles by tileeditor.py --export */', file=fp)
        scheme_lines = []
        scheme_names = []
        for selected_scheme in selection:
            idx = schemenames.index(selected_scheme)
            scheme = schemes[idx]
            line = []
            for color in scheme:
                line.append('{ 0x%02x, 0x%02x, 0x%02x }' % tuple(color))
            scheme_lines.append('{ %s }, // %s' % (', '.join(line), schemenames[idx]))
            scheme_names.append('"%s",' % (schemenames[idx],))
        tile_lines = []
        tile_names = []
        for filename in (fn for fn in existing if fn != 'tiles/tile.dat' and fn.startswith('tiles/')):
            bits = tile_file_to_2bit(filename)
            tile_lines.append('{ { %s} }, // %s' % (bits, filename[:-4]))
            tile_names.append('"%s",' % (filename[:-4],))
        digit_lines = []
        for filename in sorted(glob.glob('digits/?.dat')):
            bits = tile_file_to_2bit(filename)
            digit_lines.append('{ { %s} }, // %s' % (bits, filename[:-4]))
        print("""
struct SchemeRGB { unsigned char r, g, b; };
struct ColorScheme {
    struct SchemeRGB primary;
    struct SchemeRGB secondary;
    struct SchemeRGB bg1;
    struct SchemeRGB bg2;
};
static struct ColorScheme schemes[] = {
    %s
};

struct PackedTile16 { unsigned char pixels_2bit[%d]; };
struct PackedTile10 { unsigned char pixels_2bit[%d]; };
static struct PackedTile16 packed_tiles[] = {
    %s
};
static struct PackedTile10 packed_digits[] = {
    %s
};
        """.strip() % ('\n    '.join(scheme_lines),
                       int(blocksize*blocksize*2/8),
                       int(10*10*2/8),
                       '\n    '.join(tile_lines),
                       '\n    '.join(digit_lines),
                       ), file=fp)
        if False: print("""
static const char *scheme_names[] = {
    %s
};
static const char *tile_names[] = {
    %s
};
        """.strip() % ('\n    '.join(scheme_names),
                       '\n    '.join(tile_names),
                       ), file=fp)
    sys.exit(0)

import pygame
from pygame.locals import *

pygame.font.init()
pygame.display.init()
scr = pygame.display.set_mode((800, 600))
pygame.display.set_caption('n=next, p=prev, r=reload(undo), c=colorcycle, s=save, q=quit')

fnt = pygame.font.SysFont('Consolas', 30)

pixels = np.zeros((blocksize, blocksize), np.uint8)

pixels[0][0] = 1
pixels[0][1] = 2
pixels[0][2] = 3

def get_tile_color(idx, color=0):
    black = [0, 0, 0]
    color = schemes[coffset][color]
    light = [int(255-0.4*(255-x)) for x in color]
    darkr = [int(0.6*x) for x in color]
    return [black, color, light, darkr][idx]

def darkit(color):
    return tuple(int(x/4) for x in color)

def darkit2(color):
    return tuple(int(x/3) for x in color)

lastcol = None
while True:
    scr.fill((5, 5, 5))

    x = 10
    for color in colors:
        pygame.draw.rect(scr, color, (x, 10, 5, 5))
        x += 6

    scr.blit(fnt.render('loaded: %s, scheme: %d (%s)' % (existing[existidx], coffset, schemenames[coffset]),
        True, (255, 255, 255)), (10, 550))

    scheme = schemes[coffset]

    pygame.draw.rect(scr, scheme[0], (30, 30, 30, 30))
    pygame.draw.rect(scr, scheme[1], (30, 70, 30, 30))
    pygame.draw.rect(scr, scheme[2], (70, 30, 30, 30))
    pygame.draw.rect(scr, scheme[3], (70, 70, 30, 30))

    xoff = 220
    yoff = 50
    gscl = 30

    for y in range(blocksize):
        for x in range(blocksize):
            if is_small and (x >= 10 or y >= 10):
                continue
            pygame.draw.rect(scr, get_tile_color(pixels[y][x]), (xoff+1+x*gscl, yoff+1+y*gscl, gscl-1, gscl-1))

    bs = (10 if is_small else blocksize)
    for x in range(xoff, xoff+bs*gscl+1, gscl):
        pygame.draw.line(scr, (255, 255, 255), (x, yoff), (x, yoff+gscl*bs))

    for y in range(yoff, yoff+bs*gscl+1, gscl):
        pygame.draw.line(scr, (255, 255, 255), (xoff, y), (xoff+gscl*bs, y))

    pygame.draw.rect(scr, darkit2(scheme[2]), (0, 190, 65, 250))
    pygame.draw.rect(scr, darkit(scheme[3]), (65, 190, 65, 250))

    PATTERN = (
        (1, 0, 0),
        (0, 1, 1),
        (1, 1, 0),
        (2, 1, 1),
        (0, 2, 1),
        (1, 2, 1),
        (2, 2, 0),
        (3, 2, 0),
        (0, 3, 1),
        (1, 3, 1),
        (2, 3, 0),
        (3, 3, 0),
    )
    for xo, yo, color in PATTERN:
        for y in range(blocksize):
            for x in range(blocksize):
                if pixels[y][x] != 0:
                    scr.set_at((30+x+xo*blocksize, 200+y+yo*blocksize), get_tile_color(pixels[y][x], color))

    for idx, ch in enumerate('1234567890'):
        draw_digit(30+8*idx, 400, ch, 0)

    for idx, ch in enumerate('8934050651'):
        draw_digit(30+8*idx, 420, ch, 1)

    for xo, yo, color in PATTERN:
        for y in range(blocksize):
            for x in range(blocksize):
                if pixels[y][x] != 0:
                    for xx in (0, 1):
                        for yy in (0, 1):
                            scr.set_at((30+x*2+xx+xo*2*blocksize, 270+y*2+yy+yo*2*blocksize), get_tile_color(pixels[y][x], color))

    pygame.display.flip()

    evt = pygame.event.poll()
    if evt.type == QUIT:
        break
    elif evt.type == KEYDOWN:
        if evt.key == K_s:
            with open('tile.dat', 'wb') as fp:
                if is_small:
                    print('writing 10x10')
                    inp = pixels.tobytes()
                    tmp = bytearray()
                    for y in range(10):
                        for x in range(10):
                            tmp.append(inp[y*blocksize+x])
                    fp.write(tmp)
                else:
                    fp.write(pixels.tobytes())
                print('written tile.dat -- rename/move it to "activate" for next export')
        elif evt.key in (K_n, K_p, K_r):
            if evt.key == K_n:
                existidx = (existidx + 1) % len(existing)
            elif evt.key == K_p:
                existidx -= 1
                if existidx < 0:
                    existidx = len(existing) - 1
            with open(existing[existidx], 'rb') as fp:
                bb = fp.read()
                if len(bb) == 100:
                    is_small = True
                    tmp = []
                    for y in range(blocksize):
                        for x in range(blocksize):
                            #tmp.append(bb[int(y*10/blocksize)*10+int(x*10/blocksize)])
                            tmp.append(bb[y*10+x] if y<10 and x<10 else 0)#int(y*10/blocksize)*10+int(x*10/blocksize)])
                    bb = bytearray(tmp)
                else:
                    is_small = False
                pixels = np.frombuffer(bb, dtype=np.uint8)
                pixels.shape = (blocksize, blocksize)
                pixels = pixels.copy()
                print('loaded existing')
        elif evt.key == K_c:
            coffset = (coffset + 1) % len(schemes)
        elif evt.key == K_q:
            break
    elif evt.type == MOUSEBUTTONDOWN:
        cx, cy = evt.pos
        cx -= xoff
        cy -= yoff
        cx = int(cx/gscl)
        cy = int(cy/gscl)
        if 0<=cx<blocksize and 0<=cy<blocksize:
            pixels[cy][cx] += 1
            pixels[cy][cx] %= 4
            lastcol = pixels[cy][cx]
    elif evt.type == MOUSEBUTTONUP:
        lastcol = None
    elif evt.type == MOUSEMOTION and lastcol is not None:
        cx, cy = evt.pos
        cx -= xoff
        cy -= yoff
        cx = int(cx/gscl)
        cy = int(cy/gscl)
        if 0<=cx<blocksize and 0<=cy<blocksize:
            pixels[cy][cx] = lastcol

    time.sleep(0.01)
