# Loonies 8192 r50 for DOS and SDL platforms
# Copyright (C) 2018-2020 Thomas Perl <m@thp.io>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import mido
import oplfm
import time
import os

# Midi Drumkit
BASS_DRUM = 36
SNARE_DRUM = 38
CLOSED_HIHAT = 42
OPEN_HIHAT = 46


class ChannelMapping:
    def __init__(self, instrument, transpose=0):
        self.instrument = instrument
        self.transpose = transpose

    def note(self, note, velocity):
        self.instrument.note(note+self.transpose, velocity)


class MultiChannelMapping:
    def __init__(self, default_note, instrument_dict):
        self.default_note = default_note
        self.instrument_dict = instrument_dict

    def note(self, note, velocity):
        instrument = self.instrument_dict.get(note)
        if instrument is not None:
            instrument.note(self.default_note, velocity)


def monitor(var, addr, value):
    var.append(addr)
    var.append(value)


def convert(input_filename, output_filename, mapping, is_realtime):
    oplfm.reset()

    # Disable rhythm section for all songs at the moment
    opl_output = bytearray([0xbd, 0x00])
    mid_output = bytearray()
    oplfm.set_monitor((lambda v: lambda addr, value: monitor(v, addr, value))(opl_output))

    for msg in mido.MidiFile(input_filename):
        #print(msg)
        if is_realtime:
            time.sleep(msg.time)

        delay = int(msg.time * 1000)
        while delay > 0:
            part = min(delay, 0xff)
            # special "delay" message to be read
            opl_output.append(0xff)
            opl_output.append(part)
            mid_output.append(0xff)
            mid_output.append(part)
            delay -= part

        if msg.type == 'program_change':
            mid_output.append(0xC0 + msg.channel)
            mid_output.append(msg.program)

        if msg.type in ('note_on', 'note_off'):
            if msg.type == 'note_off':
                msg.velocity = 0

            # TODO: Only send status byte if necessary
            mid_output.append(0x90 + msg.channel)
            mid_output.append(msg.note)
            mid_output.append(msg.velocity)

            chan = mapping.get(msg.channel)
            if chan is not None:
                chan.note(msg.note, msg.velocity)

    with open(output_filename, 'wb') as fp:
        fp.write(opl_output)
        print(f'{input_filename} -> {output_filename} ({len(opl_output)} bytes)')

    with open(output_filename.replace('.out', '.dim'), 'wb') as fp:
        fp.write(mid_output)
        print(f'{input_filename} -> {output_filename.replace(".out", ".dim")} ({len(mid_output)} bytes)')

    oplfm.set_monitor(None)
    oplfm.reset()


PERCUSSION_MAPPING = MultiChannelMapping(0x30, {BASS_DRUM: oplfm.Instrument('BDRUM1'),
                                                SNARE_DRUM: oplfm.Instrument('NOISE1')})


os.makedirs('output', exist_ok=True)

CLICKEDY_MAPPING = {
    0: ChannelMapping(oplfm.Instrument('ELPIANO1')),  # Melody
    1: ChannelMapping(oplfm.Instrument('GUITAR1')),  # Bass
    9: PERCUSSION_MAPPING,  # Percussion
}
convert('input/clickedy.mid', 'output/clickedy.out', CLICKEDY_MAPPING, False)

SAMOFF_MAPPING = {
    0: ChannelMapping(oplfm.Instrument('SYN4')),  # Bass
    1: ChannelMapping(oplfm.Instrument('ELPIANO2'), -24),  # Melody
    9: PERCUSSION_MAPPING,  # Percussion
}
convert('input/samoff.mid', 'output/samoff.out', SAMOFF_MAPPING, False)

#TESTSONG_MAPPING = {
#    0: ChannelMapping(oplfm.Instrument('LOGDRUM1')),  # Melody
#    1: ChannelMapping(oplfm.Instrument('BASS1')),  # Bass
#}
#convert('input/testsong.mid', 'output/testsong.out', TESTSONG_MAPPING, False)
