/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


/* Audio abstraction: Music Player (OPL-2, MIDI (MPU/SBMIDI), CD Audio, PC Speaker) */


struct MusicTrack {
    unsigned char *data;
    unsigned int length;
};

struct MusicDef {
    struct MusicTrack data[2];
    unsigned char cda_track;
    unsigned char melody_channel;
};

/* C Header File: ADLIB *****************************************************
Author:                Kevin A. Lee
Last Amended:        27th April, 1993
Description:    Low-level interface to the Adlib (or compatible)
                FM sound card. All information gleaned from
                Jeffrey S. Lee's "Programming the Adlib/Sound
                Blaster FM Music Chips". See Lee's document for
                further information.
****************************************************************************/

void FMReset(unsigned short base)
{
   int i;

   /* zero all registers */
   for (i = 0x01; i < 0xF6; i++) WriteFM(base, i, 0);

   /* allow FM chips to control the waveform of each operator */
   WriteFM(base, 0x01, 0x20);

   /* set rhythm enabled (6 melodic voices, 5 percussive) */
   WriteFM(base, 0xBD, 0x20);
}

enum AdlibFx {
    AFX_BASS_DRUM = 0x10,
    AFX_SNARE_DRUM = 0x08,
    AFX_TOM_TOM = 0x04,
    AFX_CYMBAL = 0x02,
    AFX_HIHAT = 0x01,
};

void afx_init(unsigned short base)
{
    static unsigned char init[] = {
        0x01, 0x20,
        0x08, 0x40,
        0x2C, 0x11,
        0x2D, 0x11,
        0x31, 0x01,
        0x32, 0x04,
        0x34, 0x0C,
        0x35, 0x01,
        0x4B, 0x10,
        0x4C, 0x10,
        0x4D, 0x10,
        0x50, 0x0B,
        0x51, 0x10,
        0x52, 0x10,
        0x53, 0x10,
        0x54, 0x10,
        0x55, 0x10,
        0x6C, 0xD2,
        0x6D, 0xD2,
        0x70, 0xA8,
        0x71, 0xF7,
        0x72, 0xF7,
        0x73, 0xD6,
        0x74, 0xF8,
        0x75, 0xF5,
        0x8C, 0x74,
        0x8D, 0x74,
        0x90, 0x4C,
        0x91, 0xB5,
        0x92, 0xB5,
        0x93, 0x4F,
        0x94, 0xB5,
        0x95, 0xB5,
        0xA6, 0x57,
        0xBD, 0x20,
        0xC7, 0x01,
        0xC8, 0x01,
        0xFF,
    };

    unsigned char *cur = init;
    while (*cur != 0xFF) {
            unsigned char addr = *cur++;
            unsigned char data = *cur++;
            WriteFM(base, addr, data);
    }
}

void afx_play(unsigned short base, enum AdlibFx fx)
{
    WriteFM(base, 0xBD, 0x20);
    if (fx & AFX_BASS_DRUM) {
        WriteFM(base, 0xA7, 0x03);
        WriteFM(base, 0xA8, 0x57);
        WriteFM(base, 0xB6, 0x11);
    }
    if (fx & AFX_TOM_TOM) {
        WriteFM(base, 0xA7, 0x22);
        WriteFM(base, 0xA8, 0x6C);
        WriteFM(base, 0xB7, 0x12);
        WriteFM(base, 0xB8, 0x11);
    }
    if (fx & AFX_HIHAT) {
        WriteFM(base, 0xA7, 0x03);
        WriteFM(base, 0xA8, 0x57);
        WriteFM(base, 0xB7, 0x00);
        WriteFM(base, 0xB8, 0x00);
    }
    if (fx & AFX_CYMBAL) {
        WriteFM(base, 0xA7, 0x03);
        WriteFM(base, 0xA8, 0x57);
        WriteFM(base, 0xB7, 0x0A);
        WriteFM(base, 0xB8, 0x09);
    }
    WriteFM(base, 0xBD, 0x20 | fx);
}


struct FMPlayer {
    FMPlayer() : base(0) {}

    void init(int mode);
    void hard_reset();
    void soft_reset();
    void write(unsigned char data);

    unsigned short base;

private:
    unsigned char addr;
    unsigned char mode;
};

void
FMPlayer::init(int mode)
{
    if (base) {
        FMReset(base);
    }

    base = 0x388;

    if (mode == AUDIO_OPL_OPL2LPT_LPT1) {
        base = get_lpt_port_base(0);
    } else if (mode == AUDIO_OPL_OPL2LPT_LPT2) {
        base = get_lpt_port_base(1);
    }

    FMReset(base);
    afx_init(base);
    addr = 0;
    mode = 0;
}

void
FMPlayer::hard_reset()
{
    if (base) {
        FMReset(base);
        WriteFM(base, 0xBD, 0x00);
    }

    addr = 0;
    mode = 0;
}

void
FMPlayer::soft_reset()
{
    if (base) {
        FMReset(base);
        afx_init(base);
    }

    addr = 0;
    mode = 0;
}

void
FMPlayer::write(unsigned char data)
{
    if (!base) {
        return;
    }

    if (mode == 0) {
        addr = data;
        mode = 1;
    } else {
        WriteFM(base, addr, data);
        mode = 0;
    }
}

// Based on: http://www.piclist.com/techref/io/serial/midi/mpu.html
struct MPU {
    MPU() : base(0) {}

    void set_base_addr(unsigned short base);
    void reset_to_mpu() { if (base) send_cmd(0xFF); }
    void reset_to_uart();
    void write(unsigned char data);

private:
    unsigned char status() { return midi_in(base + 1); }
    void send_cmd(unsigned char data);

    unsigned short base;
};

void
MPU::set_base_addr(unsigned short base)
{
    if (this->base != 0) {
        // When changing the base addr, must deinit the old one
        reset_to_mpu();
    }

    this->base = base;
}

void
MPU::reset_to_uart()
{
    if (!base) {
        return;
    }

    // Reset the MPU
    send_cmd(0xFF);

    // Wait for ACK for reset cmd
    do {
        while (status() & 0x80);
    } while (midi_in(base) != 0xFE);

    // Switch to UART mode
    send_cmd(0x3F);
}

void MPU::write(unsigned char data)
{
    if (!base) {
        return;
    }

    while (status() & 0x40);
    midi_out(base, data);
}

void MPU::send_cmd(unsigned char data)
{
    while (status() & 0x40);
    midi_out(base + 1, data);
}

// https://pdos.csail.mit.edu/6.828/2011/readings/hardware/SoundBlaster.pdf
struct SBMIDI {
    SBMIDI() : base(0) {}

    void reset();
    void set_base_addr(unsigned short base);
    void write(unsigned char data);

private:
    unsigned short base;
};

void
SBMIDI::reset()
{
    if (!base) {
        return;
    }

    unsigned short reset_port = base + 0x6;
    unsigned short read_port = base + 0xA;
    unsigned short read_status_port = base + 0xE;

    midi_out(reset_port, 1);
    // delay loop?
    for (int i=0; i<255; ++i);
    midi_out(reset_port, 0);

    do {
        while (!(midi_in(read_status_port) & 0x80));
    } while (midi_in(read_port) != 0xAA);
}

void
SBMIDI::set_base_addr(unsigned short base)
{
    if (this->base) {
        reset();
    }

    this->base = base;
}

void
SBMIDI::write(unsigned char data)
{
    unsigned short write_port = base + 0xC;

    if (!base) {
        return;
    }

    while (midi_in(write_port) & 0x80);
    midi_out(write_port, 0x38);
    while (midi_in(write_port) & 0x80);
    midi_out(write_port, data);
}

struct MIDIPlayer {
    MIDIPlayer() : is_mpu(true) {}
    void init(int mode);
    void hard_reset();
    void soft_reset();
    void write(unsigned char data);

private:
    void all_notes_off()
    {
        // all notes off for all channels
        for (int channel=0; channel<16; ++channel) {
            write(0xB0 + channel);
            write(123);
            write(0);
        }
    }

    MPU mpu;
    SBMIDI sbmidi;
    bool is_mpu;
};

void
MIDIPlayer::init(int mode)
{
    unsigned short base = 0x330;
    is_mpu = true;
    if (mode == AUDIO_MPU_300) {
        base = 0x300;
    } else if (mode == AUDIO_MPU_310) {
        base = 0x310;
    } else if (mode == AUDIO_MPU_320) {
        base = 0x320;
    } else if (mode == AUDIO_MPU_SBMIDI_220) {
        base = 0x220;
        is_mpu = false;
    } else if (mode == AUDIO_MPU_SBMIDI_240) {
        base = 0x240;
        is_mpu = false;
    }

    if (is_mpu) {
        mpu.set_base_addr(base);
        mpu.reset_to_uart();
    } else {
        sbmidi.set_base_addr(base);
        sbmidi.reset();
    }
}

void
MIDIPlayer::hard_reset()
{
    all_notes_off();
    if (is_mpu) {
        mpu.reset_to_mpu();
    } else {
        sbmidi.reset();
    }
}

void
MIDIPlayer::soft_reset()
{
    all_notes_off();
    if (is_mpu) {
        mpu.reset_to_uart();
    } else {
        sbmidi.reset();
    }
}

void
MIDIPlayer::write(unsigned char data)
{
    if (is_mpu) {
        mpu.write(data);
    } else {
        sbmidi.write(data);
    }
}

struct PCSpeakerMIDI {
    PCSpeakerMIDI() : mode(0), index(0), current_note(0) {}
    void write(unsigned char data);

    unsigned char mode;
    unsigned char buf[3];
    unsigned char index;
    unsigned char melody_channel;
    unsigned char current_note;
};

void
PCSpeakerMIDI::write(unsigned char data)
{
    if ((data & 0x80) != 0) {
        index = 0;
    } else {
        ++index;

        if (index == 3) {
            // FIXME? Sticky status
            index = 1;
        }
    }

    buf[index] = data;

    if (index == 2) {
        if ((buf[0] & 0xF0) == 0x90) {
            int channel = (buf[0] & 0x0F);
            if (channel == melody_channel) {
                int note = (buf[1] & 0x7F);
                unsigned short freq = MIDIFREQ[note];
                if (buf[2] == 0) {
                    // note off
                    if (current_note == note) {
                        pcspk(false, freq);
                        current_note = 0;
                    }
                } else {
                    // note on
                    if ((current_note == 0) || current_note > note) {
                        pcspk(true, freq);
                        current_note = note;
                    }
                }
            }
        }
    }
}

struct MusicPlayer {
    MusicPlayer();
    ~MusicPlayer();

    void shutdown();
    void configure(struct HardwareConfig *hw_cfg);
    void reset(MusicDef *def);
    void update(time_ms_t delta);
    void fx(int effect);

    struct HardwareConfig *hw_cfg;
    MusicDef *def;

private:
    FMPlayer fm;
    MIDIPlayer midi;
    PCSpeakerMIDI pcspkmidi;

    unsigned char *data;
    unsigned int length;
    unsigned int pos;
    time_ms_t now;
    time_ms_t next;
};

MusicPlayer::MusicPlayer()
{
    hw_cfg = 0;
    def = 0;
    // fm, midi
    data = 0;
    length = 0;
    pos = 0;
    now = 0;
    next = 0;
}

MusicPlayer::~MusicPlayer()
{
    shutdown();
}

void
MusicPlayer::shutdown()
{
    this->data = 0;
    this->length = 0;
    this->pos = 0;
    this->now = 0;
    this->next = 0;

    if (hw_cfg) {
        switch (hw_cfg->audio_dev) {
            case AUDIO_DEV_OPL2:
                fm.hard_reset();
                break;
            case AUDIO_DEV_PC_SPEAKER:
                pcspk(0, 440);
                break;
            case AUDIO_DEV_MPU401:
                midi.hard_reset();
                break;
            case AUDIO_DEV_CDROM:
                cd_stop();
                break;
            default:
                break;
        }
    }
}

void
MusicPlayer::configure(struct HardwareConfig *hw_cfg)
{
    // TODO: Add support for reconfigure

    this->hw_cfg = hw_cfg;

    switch (hw_cfg->audio_dev) {
        case AUDIO_DEV_OPL2:
            fm.hard_reset();
            fm.init(hw_cfg->opl_setup);
            break;
        case AUDIO_DEV_PC_SPEAKER:
            pcspk(false, 440);
            break;
        case AUDIO_DEV_MPU401:
            midi.hard_reset();
            midi.init(hw_cfg->mpu_setup);
            break;
        case AUDIO_DEV_CDROM:
            cd_init();
            cd_stop();
            break;
        default:
            this->hw_cfg = 0;
            break;
    }
}

void
MusicPlayer::reset(MusicDef *def)
{
    if (!hw_cfg) {
        return;
    }

    switch (hw_cfg->audio_dev) {
        case AUDIO_DEV_OPL2:
            fm.soft_reset();
            // TODO: Disable rhythm section?
            break;
        case AUDIO_DEV_PC_SPEAKER:
            pcspk(false, 440);
            break;
        case AUDIO_DEV_MPU401:
            midi.soft_reset();
            break;
        case AUDIO_DEV_CDROM:
            cd_stop();
            cd_play_track(def->cda_track);
            break;
        default:
            break;
    }

    this->def = def;
    if (hw_cfg->audio_dev == AUDIO_DEV_OPL2 ||
            hw_cfg->audio_dev == AUDIO_DEV_MPU401 ||
            hw_cfg->audio_dev == AUDIO_DEV_PC_SPEAKER) {
        // OPL data in index 0, MIDI data in index 1
        int index = (hw_cfg->audio_dev == AUDIO_DEV_OPL2) ? 0 : 1;
        this->data = def->data[index].data;
        this->length = def->data[index].length;
        pcspkmidi.melody_channel = def->melody_channel;
        pcspkmidi.current_note = 0;
        pcspk(false, 440);
    } else {
        this->data = 0;
        this->length = 0;
    }
    this->pos = 0;
    this->now = 0;
    this->next = 0;
}

void
MusicPlayer::update(time_ms_t delta)
{
    if (!hw_cfg) {
        return;
    }

    now += delta;

    if (now < next) {
        return;
    }

    if (hw_cfg->audio_dev == AUDIO_DEV_CDROM) {
        if (next < now) {
            // allow CD implementation to process stuff
            cd_process();
            // schedule again in 100 ms
            next = now + 100;
        }
        return;
    }

    while (pos < length && next < now) {
        unsigned char cmd = data[pos++];
        if (cmd == 0xff) {
            unsigned char value = data[pos++];
            next = now + value;
        } else {
            if (hw_cfg->audio_dev == AUDIO_DEV_OPL2) {
                fm.write(cmd);
            } else if (hw_cfg->audio_dev == AUDIO_DEV_MPU401) {
                midi.write(cmd);
            } else if (hw_cfg->audio_dev == AUDIO_DEV_PC_SPEAKER) {
                pcspkmidi.write(cmd);
            }
        }

        if (pos == length) {
            pos = 0;
        }
    }
}

void
MusicPlayer::fx(int effect)
{
    if (!hw_cfg) {
        return;
    }

    switch (hw_cfg->audio_dev) {
        case AUDIO_DEV_OPL2:
            afx_play(fm.base, (enum AdlibFx)effect);
            break;
        case AUDIO_DEV_PC_SPEAKER:
            {
                int freq = 220;
                if (effect == AFX_SNARE_DRUM) {
                    freq = 330;
                }
                pcspk(true, freq);
                {
                    time_ms_t then = time_now_ms() + 10;
                    while (time_now_ms() < then);
                }
                pcspk(false, freq);
            }
            break;
        case AUDIO_DEV_MPU401:
            // Disable all rhythm notes
            midi.write(0xB9);
            midi.write(123);
            midi.write(0);

            midi.write(0x99);
            switch (effect) {
                case AFX_BASS_DRUM:
                    midi.write(36);
                    break;
                case AFX_SNARE_DRUM:
                    midi.write(38);
                    break;
                case AFX_HIHAT:
                    midi.write(42); // closed hihat!
                    break;
                case AFX_TOM_TOM:
                case AFX_CYMBAL:
                default:
                    midi.write(46); // open hihat
                    break;
            }
            midi.write(0x40);
            break;
        default:
            break;
    }
}
