/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


/* Core Game Logic for Loonies */

/* Forward-declarations of global callbacks */
void change_tileset();
void change_colors();
void change_background();

struct Block {
    void update(int a, int b, int c, int d)
    {
        this->value[0] = a;
        this->value[1] = b;
        this->value[2] = c;
        this->value[3] = d;
        y = -1;
        x = 7;
    }

    int at(int x, int y) {
        x -= this->x;
        y -= this->y;
        if ((x == 0 || x == 1) && (y == 0 || y == 1)) {
            return value[y*2+x];
        }
        return 0;
    }

    void rotate() {
        int tmp[4];
        tmp[0] = value[0];
        tmp[1] = value[1];
        tmp[2] = value[2];
        tmp[3] = value[3];
        value[0] = tmp[2];
        value[1] = tmp[0];
        value[2] = tmp[3];
        value[3] = tmp[1];
    }

    int value[4];
    int y;
    int x;
};

static struct { unsigned char a, b, c, d; } BLOCKS[] = {
    { 1, 1, 2, 2 },
    { 1, 2, 2, 1 },
    { 1, 1, 1, 2 },
    { 1, 1, 2, 2 },
    { 2, 2, 1, 1 },
    { 1, 1, 1, 1 },
    { 1, 2, 1, 2 },
    { 2, 2, 2, 1 },
    { 2, 1, 2, 1 },
    { 2, 2, 2, 2 },
    { 1, 2, 3, 2 },
    { 2, 1, 1, 2 },
    { 1, 2, 1, 2 },
};

struct Queue {
    Queue()
    {
        shuffle();
    }

    void shuffle() {
        for (int i=0; i<sizeof(BLOCKS)/sizeof(BLOCKS[0]); i++) {
            queue[i].update(BLOCKS[i].a, BLOCKS[i].b,
                            BLOCKS[i].c, BLOCKS[i].d);
        }
        index = 0;
    }

    Block *next() {
        if (index == (sizeof(BLOCKS)/sizeof(BLOCKS[0])-1)) {
            shuffle();
        }

        return &(queue[index++]);
    }

    Block queue[sizeof(BLOCKS)/sizeof(BLOCKS[0])];
    int index;
};


enum FieldBits {
    // 0, 1, 2, 3 -> value of cell
    // 0 == empty cell
    // 1 == color 1
    // 2 == color 2
    // 3 == bomb
    FIELD_VALUE_MASK = (1 << 0) | (1 << 1) | (1 << 2),
    // field value when it's a bomb
    FIELD_VALUE_BOMB = 3,
    // 1 if field is high, 0 if not
    FIELD_HIGH = (1 << 4),
    // 1 if field is queued for clear, 0 if not
    FIELD_CLEAR = (1 << 5),
    // 1 if field needs redrawing, 0 if not
    FIELD_DIRTY = (1 << 6),
    // 1 if field was high before, 0 if not
    FIELD_WAS_HIGH = (1 << 7),
    // mask of animation index (0..7)
    FIELD_ANIMATION_MASK = (1 << 8) | (1 << 9) | (1 << 10),
    // shift of animation index
    FIELD_ANIMATION_SHIFT = 8,
    // number of frames in an animation
    FIELD_ANIMATION_FRAMES = 8,
};

struct LevelDefinition {
    time_ms_t slider_interval_ms;
    time_ms_t block_interval_ms;
    unsigned short change_at_blocks;
};

static const struct LevelDefinition
LEVELS[10] = {
    { 700UL, 1500UL,  10U },  // super super easy
    { 632UL, 1245UL,  20U },
    { 504UL,  991UL,  30U },
    { 406UL,  636UL, 100U },
    { 408UL,  582UL, 350U },
    { 301UL,  427UL, 400U },
    { 203UL,  373UL, 520U },
    { 205UL,  218UL, 670U },
    { 107UL,  164UL, 800U },
    {  90UL,  110UL, 920U },  // super super hard
};

struct Field {
    Field()
        : ts_fall(40UL)
        , ts_slider(200UL)
        , ts_block(1000UL)
        , ts_animation(120UL)
    {
        width = 16;
        height = 14;
        for (int i=0; i<width*height; i++) {
            data[i] = FIELD_DIRTY;
        }
        line_at = 0;
        slider_x = 0;
        block = 0;
        force_fall = false;
        new_block();
        score = 0;
        score_display = 0;
        blocks = 0;
        quit = false;
        time = 0;

        level = 0;
        set_level(level);

        block_moved = true;
        fast_fall = false;
        fast_fall_counter = 0;
        side_movement = 0;
        side_movement_counter = 0;

        shaking_index = -1;
    }

    void set_level(int level) {
        const struct LevelDefinition *def = &LEVELS[level];
        ts_slider.set_interval(def->slider_interval_ms);
        ts_block.set_interval(def->block_interval_ms);
        this->level = level;
    }

    void next_level() {
        if (level < (sizeof(LEVELS)/sizeof(LEVELS[0]))-1) {
            ++level;
            set_level(level);
        }
    }

    void step_time() {
        time+=3;

        ts_fall.update(10);
        ts_slider.update(10);
        ts_block.update(10);
        ts_animation.update(10);

        while (ts_animation.step()) {
            for (int y=0; y<height; ++y) {
                for (int x=0; x<width; ++x) {
                    unsigned short *here = data + (y*width) + x;
                    if (((*here) & FIELD_VALUE_MASK) == FIELD_VALUE_BOMB) {
                        // update animation index
                        int animation_now = ((*here) & FIELD_ANIMATION_MASK) >> FIELD_ANIMATION_SHIFT;
                        animation_now = (animation_now + 1) % FIELD_ANIMATION_FRAMES;

                        *here &= ~FIELD_ANIMATION_MASK;

                        if (animation_now == 0) {
                            // explode the bomb
                            for (int yy=y-1; yy<=y+1; ++yy) {
                                if (yy >= 0 && yy < height) {
                                    for (int xx=x-1; xx<=x+1; ++xx) {
                                        if (xx >= 0 && xx < width) {
                                            data[yy*width+xx] &= ~(FIELD_VALUE_MASK | FIELD_HIGH | FIELD_CLEAR);
                                            data[yy*width+xx] |= FIELD_DIRTY;
                                        }
                                    }
                                }
                            }
                            shaking_index = 0;
                        } else {
                            // update the animation
                            *here |= (animation_now << FIELD_ANIMATION_SHIFT);
                        }
                    }
                }
            }
        }

        while (ts_fall.step()) step_fall();
        while (ts_slider.step()) step_slider();
        while (ts_block.step()) step_block();
    }

    void new_block() {
        end_fast_fall();

        block_dirty();
        block = queue.next();
        block_dirty();
        for (int y=block->y; y<=block->y+1; y++) {
            if (y < 0) {
                continue;
            }
            for (int x=block->x; x<=block->x+1; x++) {
                data[offset(x, y)] |= FIELD_DIRTY;
                if (at(x, y) != 0) {
                    quit = true;
                }
            }
        }
    }

    void step_slider() {
        clear_at(slider_x);

        slider_dirty();
        slider_x++;
        if (slider_x == width) {
            int total = 0;
            bool tileset_changed = false;
            bool color_changed = false;
            for (int y=0; y<height; y++) {
                for (int x=0; x<width; x++) {
                    if (data[offset(x, y)] & FIELD_CLEAR) {
                        data[offset(x, y)] &= ~(FIELD_CLEAR | FIELD_VALUE_MASK);
                        data[offset(x, y)] |= FIELD_DIRTY;
                        total += 10;
                        blocks++;

                        // level change
                        if (blocks == LEVELS[level].change_at_blocks) {
                            next_level();
                        }

                        // tileset change
                        if (blocks % 60 == 0 && !tileset_changed) {
                            change_tileset();
                            redraw_all();
                            tileset_changed = true;
                        }

                        // colors change
                        if (blocks % 80 == 0 && !color_changed) {
                            change_colors();
                            color_changed = true;
                        }
                    }
                }
            }

            // background changes if 10 are cleared in a row
            if (total > 100) {
                change_background();
            }

            if (total > 160) {
                total *= 10;
            } else if (total > 80) {
                total *= 2;
            }

            score += total;
            if (score > MAX_SCORE) {
                score = MAX_SCORE;
            }

            update_high();
            slider_x = 0;
        }
        slider_dirty();
        // also mark block dirty if slider moves over/off of it
        if (block->x == slider_x || block->x+1 == slider_x || block->x+2 == slider_x) {
            block_dirty();
        }
    }

    void clear_at(int x) {
        for (int y=0; y<height; y++) {
            if (at(x, y) && at_high(x, y)) {
                if ((data[offset(x, y)] & FIELD_CLEAR) == 0) {
                    data[offset(x, y)] |= FIELD_CLEAR | FIELD_DIRTY;
                }
            }
        }
    }

    void move_block_x(int xoff) {
        if (force_fall) {
            return;
        }

        if (block->x + xoff >= 0 && block->x + xoff < width - 1 &&
                at(block->x + xoff, block->y) == 0 &&
                at(block->x + xoff, block->y + 1) == 0 &&
                at(block->x + xoff + 1, block->y) == 0 &&
                at(block->x + xoff + 1, block->y + 1) == 0) {
            block_dirty();
            block->x += xoff;
            block_dirty();
        }
    }

    void rotate_block() {
        block->rotate();
        block_dirty();
    }

    void slider_dirty() {
        for (int y=0; y<height; y++) {
            data[offset(slider_x, y)] |= FIELD_DIRTY;
        }
        slider_moved = true;
    }

    void block_dirty() {
        int offsets[4];
        if (!block) return;

        offsets[0] = offset(block->x, block->y);
        offsets[1] = offset(block->x+1, block->y+1);
        offsets[2] = offset(block->x, block->y+1);
        offsets[3] = offset(block->x+1, block->y);

        for (int i=0; i<sizeof(offsets)/sizeof(offsets[0]); ++i) {
            if (offsets[i] != -1) {
                data[offsets[i]] |= FIELD_DIRTY;
            }
        }

        if (slider_x >= block->x-1 && slider_x <= block->x+2) {
            // redraw slider if it's over the block now
            slider_dirty();
        }

        block_moved = true;
    }

    void begin_fast_fall()
    {
        fast_fall_counter = 1;
        fast_fall = true;
    }

    void end_fast_fall()
    {
        fast_fall = false;
    }

    void begin_side_movement(int delta)
    {
        move_block_x(delta);
        side_movement_counter = -4;
        side_movement = delta;
    }

    void end_side_movement()
    {
        side_movement = 0;
    }

    void step_fall() {
        if (fast_fall) {
            ++fast_fall_counter;
            if (fast_fall_counter % 2 == 0) {
                step_block();
            }
        }

        if (side_movement) {
            ++side_movement_counter;
            if (side_movement_counter > 0 && side_movement_counter % 3 == 0) {
                move_block_x(side_movement);
            }
        }

        if (force_fall) {
            step_block();
        }

        if (score_display < score) {
            score_display += (score - score_display) / 2;
            if (score - score_display < 2) {
                score_display = score;
            }
        }

        for (int x=0; x<width; x++) {
            for (int y=height-2; y>=0; y--) {
                if (at(x, y+1) == 0 && at(x, y) != 0) {
                    set(x, y+1, at(x, y));
                    set(x, y, 0);
                }
            }
        }

        update_high();
    }

    void step_block() {
        if ((int)block->y == height - 2 ||
                        at(block->x, block->y + 2) != 0 ||
                        at(block->x+1, block->y + 2) != 0) {
            for (int x=block->x; x<=block->x+1; x++) {
                for (int y=block->y; y<=block->y+1; y++) {
                    if (at(x, y) != 0) {
                        quit = true;
                    }
                    set(x, y, block->at(x, y));
                }
            }
            new_block();
            force_fall = false;
        }
        block_dirty();
        block->y++;
        block_dirty();
        update_high();
    }

    void update_high() {
        int i;
        for (i=0; i<width*height; i++) {
            if (data[i] & FIELD_HIGH) {
                data[i] &= ~(FIELD_HIGH);
                data[i] |= FIELD_WAS_HIGH;
            } else {
                data[i] &= ~(FIELD_WAS_HIGH);
            }
        }

        for (int y=0; y<height-1; y++) {
            for (int x=0; x<width-1; x++) {
                int value = at(x, y);
                if (value == 0) {
                    continue;
                }
                if (at(x+1, y) == value && at(x+1, y+1) == value &&
                        at(x, y+1) == value) {
                    set_high(x, y, 1);
                    set_high(x+1, y, 1);
                    set_high(x, y+1, 1);
                    set_high(x+1, y+1, 1);
                }
            }
        }
        for (i=0; i<width*height; i++) {
            // all the cells that lost their "high" status need a dirty flag
            if ((data[i] & FIELD_WAS_HIGH) && !(data[i] & FIELD_HIGH)) {
                data[i] |= FIELD_DIRTY;
                data[i] &= ~(FIELD_WAS_HIGH);
            }
        }

    }

    int offset(int x, int y) {
        if (x >= 0 && x < width && y >= 0 && y < height) {
            return y * width + x;
        } else {
            return -1;
        }
    }

    void set(int x, int y, int value) {
        if ((data[offset(x, y)] & FIELD_VALUE_MASK) != value) {
            data[offset(x, y)] &= ~(FIELD_VALUE_MASK);
            data[offset(x, y)] |= (value & FIELD_VALUE_MASK) | FIELD_DIRTY;
        }
    }

    void redraw_all() {
        for (int i=0; i<width*height; ++i) {
            data[i] |= FIELD_DIRTY;
        }
        block_moved = true;
        slider_moved = true;
    }

    void set_high(int x, int y, int value) {
        int old_value = ((data[offset(x, y)] & FIELD_WAS_HIGH) != 0);

        if (value) {
            data[offset(x, y)] |= FIELD_HIGH;
        } else {
            data[offset(x, y)] &= ~(FIELD_HIGH);
        }

        if (value != old_value) {
            data[offset(x, y)] |= FIELD_DIRTY;
        }
    }

    int at(int x, int y) {
        return data[offset(x, y)] & FIELD_VALUE_MASK;
    }

    int at_high(int x, int y) {
        return (data[offset(x, y)] & FIELD_HIGH) != 0;
    }

    bool cells(void (*func)(int x, int y, int color)) {
        if (block_moved) {
            block_moved = false;
            for (char y=block->y; y<block->y+2; ++y) {
                if (y<0 || y>=height) {
                    continue;
                }
                for (char x=block->x; x<block->x+2; ++x) {
                    if (x<0 || x>=width) {
                        continue;
                    }
                    unsigned char v = block->at(x, y);
                    if ((v & FIELD_VALUE_MASK) == FIELD_VALUE_BOMB) {
                        v = 7 + ((v & FIELD_ANIMATION_MASK) >> FIELD_ANIMATION_SHIFT);
                    }
                    if (v) {
                        func(x, y, v);
                        if (v < 7) {
                            data[y*width+x] &= ~(FIELD_DIRTY);
                        } else {
                            block_moved = true;
                        }
                    }
                }
            }
        }

        unsigned short *here = data;
        for (int y=0; y<height; y++) {
            for (int x=0; x<width; x++) {
                unsigned short v = *here;
                if (v & FIELD_DIRTY) {
                    if (v & FIELD_CLEAR) {
                        v &= FIELD_VALUE_MASK;
                        v += 4;
                    } else if (v & FIELD_HIGH) {
                        v &= FIELD_VALUE_MASK;
                        v += 2;
                    } else {
                        v &= FIELD_VALUE_MASK;
                        if (v == FIELD_VALUE_BOMB) {
                            v = 7 + (((*here) & FIELD_ANIMATION_MASK) >> FIELD_ANIMATION_SHIFT);
                        }
                    }
                    func(x, y, v);
                    if (v < 7) {
                        *here &= ~(FIELD_DIRTY);
                    }
                }

                ++here;
            }
        }

        if (slider_moved) {
            slider_moved = false;
            return true;
        }

        return false;
    }

    Timestep ts_fall;
    Timestep ts_slider;
    Timestep ts_block;
    Timestep ts_animation;

    int width;
    int height;
    unsigned short data[16*14];
    int line_at;
    int slider_x;
    Block *block;
    bool force_fall;
    Queue queue;
    score_t score;
    score_t score_display;
    score_t blocks;
    bool quit;
    int time;
    score_t level;
    bool block_moved;
    bool slider_moved;
    bool fast_fall;
    int fast_fall_counter;
    int side_movement;
    int side_movement_counter;
    int shaking_index;
};

