/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


#include "loonport.h"


#include <SDL.h>
#include <vector>
#include <thread>
#include <map>
#include <cmath>
#include <unistd.h>

#include "softvga.h"


// Define this to zero to see all offscreen data
#if 1
#define VISIBLE_HEIGHT PORT_HEIGHT
#else
#define VISIBLE_HEIGHT (4*PORT_HEIGHT)
#endif

SDL_Surface *screen;
SDL_Window *window;

static std::map<int,int>
KEYCODE_MAP = {
    {SDLK_LEFT, KeyCode_Left},
    {SDLK_RIGHT, KeyCode_Right},
    {SDLK_DOWN, KeyCode_Down},
    {SDLK_UP, KeyCode_Up},
    {SDLK_RETURN, KeyCode_Enter},
    {SDLK_ESCAPE, KeyCode_Escape},
    {SDLK_SPACE, KeyCode_Space},
    {SDLK_q, KeyCode_q},
    {SDLK_h, KeyCode_h},
    {SDLK_j, KeyCode_j},
    {SDLK_k, KeyCode_k},
    {SDLK_l, KeyCode_l},
    {SDLK_y, KeyCode_y},
    {SDLK_n, KeyCode_n},
    {SDLK_p, KeyCode_p},
};


bool _map_key(PortEvent *event, int keysym)
{
    auto it = KEYCODE_MAP.find(keysym);
    if (it != KEYCODE_MAP.end()) {
        event->key = it->second;
        return true;
    } else {
        return false;
    }
}

static bool
_event_proc(int (*func)(SDL_Event *), PortEvent *event)
{
    SDL_Event e;
    int res = func(&e);
    if (!res) {
        return false;
    }

    if (e.type == SDL_QUIT) {
        event->type = EVENT_QUIT;
    } else if (e.type == SDL_KEYDOWN) {
        event->type = EVENT_KEYDOWN;
        return _map_key(event, e.key.keysym.sym);
    } else if (e.type == SDL_KEYUP) {
        event->type = EVENT_KEYUP;
        return _map_key(event, e.key.keysym.sym);
    } else {
        return false;
    }

    return true;
}

bool event_poll(PortEvent *event)
{
    return _event_proc(SDL_PollEvent, event);
}

bool event_wait(PortEvent *event)
{
    while (!_event_proc(SDL_WaitEvent, event)) {}
    return true;
}


bool
save_file(const char *filename, void *data, unsigned short length)
{
    bool written = true;
    FILE *fp = fopen(filename, "wb");
    if (!fp) {
        return false;
    }
    written = (fwrite(data, length, 1, fp) == 1);
    fclose(fp);
    return written;
}

bool
load_file(const char *filename, void *data, unsigned short len)
{
    bool success = true;
    FILE *fp = fopen(filename, "rb");
    if (!fp) {
        return false;
    }
    success = (fread(data, len, 1, fp) == 1);
    fclose(fp);
    return success;
}

void WriteFM(unsigned short port, uint8_t reg, uint8_t value)
{
    // TODO: Add support for OPL2LPT
    (void)port;
    (void)reg;
    (void)value;
}

time_ms_t time_now_ms()
{
    return SDL_GetTicks();
}


static bool pcspk_enabled = false;
static int16_t pcspk_frequency = 440;

void audioCallback(void *buf, unsigned int length, void *userdata);

void MyAudioCallback(void*  userdata,
                       Uint8* stream,
                       int    len)
{
    audioCallback(stream, len/4, userdata);
}



void vgamode(enum VGAMode mode)
{
    if (mode == VGAMode_320x200 || mode == VGAMode_320x240) {
        SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);

        SDL_AudioSpec want, have;
        SDL_AudioDeviceID dev;

        SDL_memset(&want, 0, sizeof(want)); /* or SDL_zero(want) */
        want.freq = 44100;
        want.format = AUDIO_S16;
        want.channels = 2;
        want.samples = 512;
        want.callback = MyAudioCallback;  // you wrote this function elsewhere.
        dev = SDL_OpenAudioDevice(NULL, 0, &want, &have, SDL_AUDIO_ALLOW_FORMAT_CHANGE);
        SDL_PauseAudioDevice(dev, 0);

        window = SDL_CreateWindow("Loonies 8192", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                PORT_WIDTH, VISIBLE_HEIGHT, SDL_WINDOW_SHOWN | SDL_WINDOW_ALLOW_HIGHDPI);
        RGB565 rmask; rmask.v = 0; rmask.r = 0b11111;
        RGB565 gmask; gmask.v = 0; gmask.g = 0b111111;
        RGB565 bmask; bmask.v = 0; bmask.b = 0b11111;
        screen = SDL_CreateRGBSurface(0, PORT_WIDTH, VISIBLE_HEIGHT, 16, rmask.v, gmask.v, bmask.v, 0x0000);
    }

    if (mode == VGAMode_Text) {
        SDL_Quit();
    }
}

void vgawait()
{
    SDL_LockSurface(screen);
    uint16_t *dst = ((uint16_t *)screen->pixels);
    for (int y=0; y<VISIBLE_HEIGHT; y++) {
        auto src = vga_scanout_start() + y * PORT_WIDTH;
        for (int x=0; x<PORT_WIDTH; x++) {
            *dst++ = PAL[*src++].v;
        }
    }
    SDL_UnlockSurface(screen);

    SDL_BlitSurface(screen, nullptr, SDL_GetWindowSurface(window), nullptr);
    SDL_UpdateWindowSurface(window);
    SDL_PumpEvents();
}

static bool
midi_inited = false;

void midi_out(unsigned short port, unsigned char data)
{
    if ((port & 1) != 0) {
        if (data == 0x3F) {
            // UART mode
            midi_inited = true;
        } else if (data == 0xFF) {
            // MPU mode
            midi_inited = false;
        }
    }

    if ((port & 1) == 0 && midi_inited) {
        // Repacketizing of MIDI events
        static unsigned char status = 0;
        static int offset = 0;
        static unsigned char buf[2];
        if (data & 0x80) {
            status = data;
            offset = 0;
        } else {
            buf[offset++] = data;
            if (offset == 2) {
                // TODO: Send midi event buf[0] + buf[1]
                offset = 0;
            }
        }
    }
}

unsigned char midi_in(unsigned short port)
{
    // from status port
    if (port & 1) {
        // ready for input/output (all flags cleared)
        return 0x00;
    }

    // ACK (from data port)
    return 0xFE;
}

unsigned short get_lpt_port_base(unsigned char index)
{
    // Does not matter on SDL yet (TODO: Add support)
    return index ? 0 : 0xabc;
}

static std::thread thr;
static bool thr_quit = true;

void
timer_init()
{
}

void
timer_quit()
{
    if (!thr_quit) {
        thr_quit = true;
        thr.join();
    }
}

void
set_timer_callback(timer_callback_t callback)
{
    thr_quit = false;
    thr = std::thread([=] () {
        auto last = time_now_ms();
        while (!thr_quit) {
            auto now = time_now_ms();
            while (now > last) {
                callback();
                ++last;
            }
            usleep(100);
        }
    });
}

bool cd_available()
{
    FILE *fp = fopen("track2.wav", "rb");
    if (fp != nullptr) {
        fclose(fp);
        return true;
    }

    return false;
}

struct AudioContext {
    FILE *fp;
    int track;
    int next_track;
} audio_context = { nullptr, 0, 0 };

void cd_init()
{
}

void cd_play_track(int track)
{
    audio_context.next_track = track;
}

void cd_process()
{
}

void cd_stop()
{
    audio_context.next_track = 0;
}

int WAV_HEADER_SIZE = 12 /* RIFF */ + 24 /* fmt_ */ + 8 /* data */;

void
audioCallback(void *buf, unsigned int length, void *userdata)
{
    if (audio_context.next_track != audio_context.track) {
        audio_context.track = audio_context.next_track;
        if (audio_context.fp != nullptr) {
            fclose(audio_context.fp);
            audio_context.fp = nullptr;
        }

        if (audio_context.track != 0) {
            char tmp[32];
            sprintf(tmp, "track%d.wav", audio_context.track);
            audio_context.fp = fopen(tmp, "rb");
            if (audio_context.fp) {
                fseek(audio_context.fp, WAV_HEADER_SIZE, SEEK_SET);
            }
        }
    }

    if (audio_context.fp == nullptr) {
        memset(buf, 0, length*4);
    } else {
        if (feof(audio_context.fp)) {
            fseek(audio_context.fp, WAV_HEADER_SIZE, SEEK_SET);
        }
        fread(buf, length*4, 1, audio_context.fp);
    }
}

void pcspk(unsigned char enabled, unsigned short frequency)
{
    pcspk_enabled = enabled;
    pcspk_frequency = frequency;
}

#include "loonies.cpp"


int
main(int, char *[])
{
    return loonies_main();
}
