/*
    Loonies 8192 r50 for DOS and SDL platforms
    Copyright (C) 2018-2020 Thomas Perl <m@thp.io>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

/**
 * Lazy implementation of VGA-like semantics (and functions)
 * for non-VGA platforms (used in SDL port of Loonies 8192)
 **/


#include <cstdio>
#include <cstdlib>
#include <cstdint>
#include <cstring>
#include <unistd.h>

#include "loonport.h"

namespace {

unsigned char _vga[PORT_WIDTH*PORT_HEIGHT*4*2];

int _vgaoffset = 0;

#if defined(BUILD_FOR_3DS)
union RGB565 { struct { uint16_t b:5, g:6, r:5; }; uint16_t v; } PAL[256];
#else
union RGB565 { struct { uint16_t r:5, g:6, b:5; }; uint16_t v; } PAL[256];
#endif

inline unsigned char *vga_scanout_start() { return _vga + _vgaoffset; }

};

void putpixel(unsigned short x, unsigned short y, uint8_t color)
{
    _vga[y*PORT_WIDTH+x] = color;
}

unsigned char getpixel(unsigned short x, unsigned short y)
{
    return _vga[y*PORT_WIDTH+x];
}

void vgascroll(unsigned int y)
{
    _vgaoffset = y * PORT_WIDTH;
}

void vgablit(unsigned short sx, unsigned short sy, unsigned short dx, unsigned short dy, unsigned short w, unsigned short h, unsigned char orvalue)
{
    //printf("vgablit(%d vs %d)\n", sx%4, dx%4);
    if (sx % 4 != dx % 4) {
        fprintf(stderr, "vgablit(sx=%d, sy=%d, dx=%d, dy=%d, w=%d, h=%d, orvalue=%d)\n",
                sx, sy, dx, dy, w, h, orvalue);
        exit(1);
    }

    for (int y=0; y<h; ++y) {
        unsigned char *src = _vga + (sy+y) * PORT_WIDTH + sx;
        unsigned char *dst = _vga + (dy+y) * PORT_WIDTH + dx;
        for (int x=0; x<w; ++x) {
            *dst++ = (*src++) | orvalue;
        }
    }
}

void render_rect(int x, int y, int w, int h, uint8_t color)
{
    for (int yy=0; yy<h; ++yy) {
        for (int i=0; i<w; ++i) {
            putpixel(x+i, y+yy, color);
        }
    }
}

void render_vline(int x, int y, int h, uint8_t color)
{
    for (int i=0; i<h; ++i) {
        putpixel(x, y+i, color);
    }
}

void vgablitmemline(int x, int y, const unsigned char *data, int w)
{
    for (int i=0; i<w; ++i) {
        putpixel(x+i, y, data[i]);
    }
}

void set_ega_color(unsigned char color, unsigned char r, unsigned char g, unsigned char b)
{
    // Data comes in as 6-bit values (as in VGA specs)
    PAL[color].r = r >> 1;
    PAL[color].g = g;
    PAL[color].b = b >> 1;
}

void vgaclearbytes_y(unsigned short y, long bytes)
{
    memset(_vga + y * PORT_WIDTH, 0, bytes);
}

void vgaclearbytes(unsigned long bytes)
{
    vgaclearbytes_y(0, bytes);
}
